import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.cm as cm
from ..hector.magnets.circular import is_circular_magnet
from ..hector.magnets.rectangular import is_rectangular_magnet
from ..hector.constants import robot_center_x,robot_center_y
import re
import numpy as np
from math import pi, cos, sin
import pandas as pd
from ..general_operations.trigonometry import rotational_matrix,convert_degrees_to_radians, convert_radians_to_degrees

plt.rc('font', size=30)          # controls default text sizes
plt.rc('axes', titlesize=30)     # fontsize of the axes title
plt.rc('axes', labelsize=30)     # fontsize of the x and y labels
plt.rc('xtick', labelsize=30)    # fontsize of the tick labels
plt.rc('ytick', labelsize=30)    # fontsize of the tick labels
plt.rc('legend', fontsize=30)    # legend fontsize
plt.rc('figure', titlesize=30)   # fontsize of the figure title


def annotate_text_in_plot(magnet):
    rotation_matrix = rotational_matrix(convert_degrees_to_radians(magnet.orientation))

    if magnet.hexabundle == 'A' or magnet.hexabundle == 'B':
        corner_offset = 9
        width = 22.4
        length = 54.4
        radius = 14
        z_order = 2
    elif magnet.hexabundle == 'C':
        corner_offset = 6
        width = 19.4
        length = 51.4
        radius = 12
        z_order = 2
    elif magnet.hexabundle == 'H' or magnet.hexabundle == 'U':
        corner_offset = 2
        width = 13.4
        length = 45.4
        radius = 9
        z_order = 3
    else:
        corner_offset = 3.5
        width = 16.4
        length = 48.4
        radius = 10
        z_order = 3

    magnet.corner = (magnet.center[0] + rotation_matrix[0][0] * (magnet.width + corner_offset) / 2 + rotation_matrix[1][0] * (magnet.length + 3) / 2, \
                 magnet.center[1] + rotation_matrix[0][1] * (magnet.width + corner_offset) / 2 + rotation_matrix[1][1] * (magnet.length + 3) / 2)
    magnet.circular_center = (magnet.center[0] - rotation_matrix[1][0] * (magnet.length + 27) / 2, \
                 magnet.center[1] - rotation_matrix[1][1] * (magnet.length + 27) / 2)
    annotated_text = re.split('(\d+)', magnet.magnet_label)[1]
    annotated_text = str(magnet.hexabundle + '  ' + str(annotated_text))

    if (annotated_text[1] == 'S'):
        annotated_text = annotated_text.replace('GS', '')
        label_color = 'black'# change back to 'yellow'
        patch_color = 'black'
    elif ('I'> magnet.hexabundle >='A'):
        label_color = 'black'
        patch_color = 'white'
    elif (magnet.hexabundle >'H'):
        label_color = 'black'# change back to 'white'
        patch_color = 'black'
    ## Increase zorder of the smaller hexabundles over A and B and C- the largest ones
    #****** previousOne draw_fill = patches.Rectangle((magnet.corner[0], magnet.corner[1]), width, length, angle=abs(180 - magnet.orientation),facecolor=patch_color,alpha=0.8,zorder=3)#,edgecolor='black',linewidth=0.5
    # draw_fill2 = patches.Rectangle((magnet.corner[0], magnet.corner[1]), width, (length-12), angle=abs(180 - magnet.orientation), facecolor=patch_color,edgecolor='black',linewidth=0.5, alpha=0.8,zorder=z_order)
    # draw_fill3 = patches.Circle((magnet.circular_center[0], magnet.circular_center[1]), radius=radius, facecolor=patch_color, edgecolor='black',linewidth=0.5, alpha=0.8,zorder=z_order)
    # draw_fill4 = patches.Rectangle((magnet.corner[0], magnet.corner[1]), width, (length - 11), angle=abs(180 - magnet.orientation),
    #                                facecolor=patch_color, linewidth=0.5, alpha=0.9,zorder=3)
    # plt.gcf().gca().add_artist(draw_fill) ## previous one
    # plt.gcf().gca().add_artist(draw_fill2)
    # plt.gcf().gca().add_artist(draw_fill3)
    # plt.gcf().gca().add_artist(draw_fill4)

    magnet.text_center = (magnet.center[0] - rotation_matrix[1][0] * (magnet.length ) / 2, \
                    magnet.center[1] - rotation_matrix[1][1] * (magnet.length ) / 2)

    if (magnet.hexabundle == 'H') or (magnet.hexabundle == 'U'):
        annotated_star = str('      ★')
        plt.annotate(annotated_star, (magnet.text_center[0], magnet.text_center[1]), color=label_color,rotation=abs(90 - magnet.orientation), \
                     weight='bold', fontsize=11, ha='center', va='center', zorder=3)

    plt.annotate(annotated_text, (magnet.text_center[0], magnet.text_center[1]), color=label_color, rotation=abs(90 - magnet.orientation),\
                 weight='bold', fontsize=6, ha='center', va='center',zorder=4)

def coordinates_and_angle_of_skyFibres(angle,radii):
    x = y = rotation = 0
    if angle>180:
        angle = -(360-angle)

    if 0<=angle<=90:
        x = -radii*cos(convert_degrees_to_radians(90-angle))
        y = radii*sin(convert_degrees_to_radians(90-angle))
        rotation = angle
    elif 90<angle<=180:
        x = -radii*cos(convert_degrees_to_radians(angle-90))
        y = -radii*sin(convert_degrees_to_radians(angle-90))
        rotation = angle
    elif -90<=angle<0:
        x = radii*cos(convert_degrees_to_radians(angle+90))
        y = radii*sin(convert_degrees_to_radians(angle+90))
        rotation = angle
    elif -180<=angle<-90:
        x = radii*sin(convert_degrees_to_radians(angle+180))
        y = -radii*cos(convert_degrees_to_radians(angle+180))
        rotation = angle

    return x,y,rotation

def read_sky_fibre_file(filename):

    dataframe = pd.read_csv(filename, header=0)
    dataframe.dropna(how="all", inplace=True)
    print(dataframe)

    subplate_name = dataframe['SUBPLATE_name']
    fibre_number = dataframe['Fibre_number']
    position = dataframe['Position']

    skyfibreDict = {}
    for i in range(1,len(subplate_name)):
        if str(subplate_name[i]) not in skyfibreDict:
            skyfibreDict[str(subplate_name[i])] = []
        skyfibreDict[str(subplate_name[i])].append({int(fibre_number[i]): int(position[i])})

    return skyfibreDict

def sky_fibre_annotations():
    angle_subplate = [7,5,3,1,-1,-3,-5,-7]
    radii = 270

    filename = 'Sky_fibre_position_example.csv'
    skyfibreDict = read_sky_fibre_file(filename)
    print(skyfibreDict)
    string = str(skyfibreDict['H3'][5].keys())
    print(re.sub('[^0-9]', '', string))

    #sky fibres top batch
    angle = 30
    skyfibreTitles_top = ['H3','A3','H4','A4']
    for i in range(0,4):
        x, y, rotation = coordinates_and_angle_of_skyFibres(angle,radii)
        plt.annotate(skyfibreTitles_top[i], (x,y), color='black',rotation=rotation, fontsize=11, ha='center', va='center')
        for j in range(len(skyfibreDict[skyfibreTitles_top[i]])):
            angle_pos = angle + angle_subplate[j]
            x, y, rotation = coordinates_and_angle_of_skyFibres(angle_pos,342)
            fibre_num = re.sub('[^0-9]', '', str(skyfibreDict[skyfibreTitles_top[i]][j].keys()))
            plt.annotate(fibre_num, (x, y), color='black', rotation=rotation, fontsize=7,ha='center', va='center')
            x, y, rotation = coordinates_and_angle_of_skyFibres(angle_pos, 308)
            plt.annotate('▮', (x, y), color='black', rotation=rotation, fontsize=7,ha='center', va='center')
            x, y, rotation = coordinates_and_angle_of_skyFibres(angle_pos, 322)
            plt.annotate(str(skyfibreDict[skyfibreTitles_top[i]][j][j+1]), (x, y), color='black', rotation=rotation, fontsize=7, ha='center',va='center')
        if skyfibreTitles_top[i][0] == 'H':
            alpha = 0.4
        elif skyfibreTitles_top[i][0] == 'A':
            alpha = 0.7
        draw_wedge = patches.Wedge((0, 0), r=333, theta1=angle-9+90, theta2=angle+9+90, width=71, facecolor='gray', edgecolor='black',alpha=alpha)
        plt.gcf().gca().add_artist(draw_wedge)
        angle = angle - 20

    # sky fibres left batch
    angle = 150
    skyfibreTitles_left = ['A1', 'H1', 'H2', 'A2']
    for i in range(0, 4):
        x, y, rotation = coordinates_and_angle_of_skyFibres(angle,radii)
        plt.annotate(skyfibreTitles_left[i], (x, y), color='black', rotation=rotation, fontsize=11,ha='center', va='center')
        for j in range(len(skyfibreDict[skyfibreTitles_left[i]])):
            angle_pos = angle + angle_subplate[j]
            x, y, rotation = coordinates_and_angle_of_skyFibres(angle_pos,342)
            fibre_num = re.sub('[^0-9]', '', str(skyfibreDict[skyfibreTitles_left[i]][j].keys()))
            plt.annotate(fibre_num, (x, y), color='black', rotation=rotation, fontsize=7, ha='center',va='center')
            x, y, rotation = coordinates_and_angle_of_skyFibres(angle_pos, 308)
            plt.annotate('▮', (x, y), color='black', rotation=rotation, fontsize=7, ha='center',va='center')
            x, y, rotation = coordinates_and_angle_of_skyFibres(angle_pos, 322)
            plt.annotate(str(skyfibreDict[skyfibreTitles_left[i]][j][j + 1]), (x, y), color='black', rotation=rotation, fontsize=7, ha='center', va='center')
        if skyfibreTitles_top[i][0] == 'H':
            alpha = 0.4
        elif skyfibreTitles_top[i][0] == 'A':
            alpha = 0.7
        draw_wedge = patches.Wedge((0, 0), r=333, theta1=angle-9+90, theta2=angle+9+90, width=71, facecolor='gray', edgecolor='black',alpha=alpha)
        plt.gcf().gca().add_artist(draw_wedge)
        angle = angle - 20

    # sky fibres right batch
    angle = -160
    skyfibreTitles_right = ['H5', 'H6', 'A5', 'H7']
    for i in range(0, 4):
        x, y, rotation = coordinates_and_angle_of_skyFibres(angle,radii)
        plt.annotate(skyfibreTitles_right[i], (x, y), color='black', rotation=rotation, fontsize=11, ha='center', va='center')
        for j in range(len(skyfibreDict[skyfibreTitles_right[i]])):
            angle_pos = angle + angle_subplate[j]
            x, y, rotation = coordinates_and_angle_of_skyFibres(angle_pos, 342)
            fibre_num = re.sub('[^0-9]', '', str(skyfibreDict[skyfibreTitles_right[i]][j].keys()))
            plt.annotate(fibre_num, (x, y), color='black', rotation=rotation, fontsize=7, ha='center',va='center')
            x, y, rotation = coordinates_and_angle_of_skyFibres(angle_pos, 308)
            plt.annotate('▮', (x, y), color='black', rotation=rotation, fontsize=7, ha='center',va='center')
            x, y, rotation = coordinates_and_angle_of_skyFibres(angle_pos, 322)
            plt.annotate(str(skyfibreDict[skyfibreTitles_right[i]][j][j + 1]), (x, y), color='black', rotation=rotation, fontsize=7, ha='center', va='center')
        if skyfibreTitles_top[i][0] == 'H':
            alpha = 0.4
        elif skyfibreTitles_top[i][0] == 'A':
            alpha = 0.7
        draw_wedge = patches.Wedge((0, 0), r=333, theta1=angle-9+90, theta2=angle+9+90, width=71, facecolor='gray', edgecolor='black',alpha=alpha)
        plt.gcf().gca().add_artist(draw_wedge)
        angle = angle + 20


def draw_all_magnets(magnets,clusterNum,tileNum,figureFile):

    plt.figure(1)
    # draw_circle = plt.Circle((0, 0), 260, fill=True, color='#E78BE7', alpha=0.4)
    # draw_circle1 = plt.Circle((0, 0), 213.98, fill=True, color='#f6f93b')
    # draw_circle2 = plt.Circle((0, 0), 163.02, fill=True, color='#60fb3d')
    # draw_circle3 = plt.Circle((0, 0), 102.96, fill=True, color='#add8e6')
    # plt.gcf().gca().add_artist(draw_circle)
    # plt.gcf().gca().add_artist(draw_circle1)
    # plt.gcf().gca().add_artist(draw_circle2)
    # plt.gcf().gca().add_artist(draw_circle3)
    #
    # sky_fibre_annotations()

    for magnet in magnets:

        if is_circular_magnet(magnet):
            plt.figure(1)
            magnet.draw_circle('r') ## The magnets being drawn
###################################################
            plt.figure(2)
            magnet.draw_circle('r')
            #plt.text(magnet.center[0], magnet.center[1], str(int(magnet.index)), fontsize=12, rotation=abs(135-magnet.orientation))
###################################################
        if is_rectangular_magnet(magnet):
            plt.figure(1)
            magnet.draw_rectangle('r') ## The magnets being drawn
            # plt.text(magnet.center[0], magnet.center[1], str(magnet.magnet_label), fontsize=9, rotation=abs(180-magnet.orientation))  #+'\n'+str(magnet.hexabundle)
            # annotate_text_in_plot(magnet)

            plt.xlabel('mm')
            plt.ylabel('mm')
################################################################
            plt.figure(2)
            magnet.draw_rectangle('r')
            plt.text(robot_center_x + magnet.center[0], robot_center_y + magnet.center[1], str(magnet.magnet_label), fontsize=8, \
                     rotation=abs(180 - magnet.orientation),weight='bold')  # +'\n'+str(magnet.hexabundle)

            plt.xlabel('mm')
            plt.ylabel('mm')
################################################################

    # plt.figure(1)

    # plt.axis('off')
    # axes = plt.gca()
    # axes.set_xlim([-350, 350])
    # axes.set_ylim([-350, 350])
    # plt.show()                ## for showing the figure of magnets with pickup area
    # plt.savefig("image.png", bbox_inches='tight', dpi=100)
    plt.savefig(figureFile)

    f = plt.figure(1)
    f1 = plt.figure(2)
    f.clear()
    f1.clear()
    plt.close(f)
    plt.close(f1)

def create_magnet_pickup_areas(magnets):

    for magnet in magnets:
        magnet.create_pickup_areas()

def draw_magnet_pickup_areas(magnets, colour):

    for magnet in magnets:
        for pickup_area in magnet.pickup_areas:
            pickup_area.draw_rectangle(colour)




import numpy as np
import string
import sys
import re

def overall_hexabundle_size_allocation_operation_version1(all_magnets, galaxyIDrecord, clusterNum, tileNum, flagsFile):

    hexabundleIndexAAOmega = 0
    hexabundleIndexHector = 0
    hexabundleIndexRemaining = 0

    answer = []
    answer2 = []
    answer3 = []

    hexabundleRepeatwithsameID = []

    hexabundleDict = {'A':'A169','B':'A169','C':'A127','D':'A91','E':'A61','F':'A61','G':'A61','H':'A37','I':'S91','J':'S61','K':'S61','L':'S61','M':'S61','N':'S61','O':'S61','P':'S61','Q':'S61','R':'S61','S':'S61','T':'S61','U':'S37'}

    for hexabundleLetter in hexabundleDict:
        hexabundleDict[hexabundleLetter] = re.split('(\d+)',hexabundleDict[hexabundleLetter])[0:2]

    # Creating magnet dictionary with all necessary parameters required for hexabundle allocation
    MagnetDict = create_Magnet_Dictionary(all_magnets)

    for i in all_magnets:

        ## Check for match of galaxies with same ID
        for recordID in galaxyIDrecord:

            # searching each magnet ID in the records
            if i.IDs == recordID and i.__class__.__name__ == 'rectangular_magnet' \
            and galaxyIDrecord[recordID] != 'U' and galaxyIDrecord[recordID] != 'H':
                print('Galaxy ID match found with previous tiles !')
                # print(recordID)
                # print(galaxyIDrecord[recordID])

                # Galaxy ID match found in the records
                if galaxyIDrecord[recordID] in hexabundleRepeatwithsameID:

                    # storing hexabundle letter in the 'case' variable
                    case = galaxyIDrecord[recordID]

                    # Allocating the next largest hexabundle when a repeat of same hexabundle allocation occurred
                    while (case in hexabundleRepeatwithsameID):
                        # AAOmega hexabundle repeat- shifting to next largest same type of bundle
                        if 'A' < case <= 'G':
                            case = chr(ord(case)-1)
                        # Spector hexabundle repeat- shifting to next largest same type of bundle
                        elif 'I' <= case < 'T':
                            case = chr(ord(case)+1)
                        # All AAOmega hexabundles allocated, so shifting to Spector and recording as flag
                        elif hexabundleIndexAAOmega == 7:
                            case = 'J'
                            with open(flagsFile, 'a') as fp:
                                writeToFile = 'ID CHECK. All AAOmega  hexabundles allocated, so shifting to Spector. Cluster/Batch :' + str(clusterNum) + ',tile ' + str(tileNum) + '; ' \
                                              + i.__class__.__name__+' '+str(int(i.index))+', galaxy ID:'+str(recordID) \
                                              + ', hexabundle: ' + str(galaxyIDrecord[recordID]) + '\n'
                                fp.write(writeToFile)
                            print('ALL AAOMEGA HAS BEEN ALLOCATED SO HECTOR WILL BE SELECTED NOW')
                        # All Spector hexabundles allocated, so shifting to AAOmega and recording as flag
                        elif hexabundleIndexHector == 12:
                            case = 'G'
                            with open(flagsFile, 'a') as fp:
                                writeToFile = 'ID CHECK. All Spector hexabundles allocated, so shifting to AAOmega. Cluster/Batch :'+str(clusterNum)+',tile '+str(tileNum)+'; ' \
                                              + i.__class__.__name__+' '+str(int(i.index))+', galaxy ID:'+str(recordID) \
                                              + ', hexabundle: '+ str(galaxyIDrecord[recordID]) + '\n'
                                fp.write(writeToFile)
                            print('ALL SPECTOR HAS BEEN ALLOCATED SO AAOMEGA WILL BE SELECTED NOW')
                        # Move back to smallest AAOmega hexabundle in case of A and B both being in records already
                        elif case == 'A':
                            if ('B' in hexabundleRepeatwithsameID):
                                case = 'G'
                            elif  ('B' not in hexabundleRepeatwithsameID):
                                case = 'B'
                        # Move back to first Spector hexabundle in case of last one being in records already
                        elif case == 'T':
                            case = 'I'
                        else:
                        # Record the exceptional case of an unresolved repeat as a flag
                            with open(flagsFile, 'a') as fp:
                                writeToFile = 'EXCEPTIONAL CASE OF UNRESOLVED REPEAT !!! \n'
                                fp.write(writeToFile)
                            break

                    # Galaxy with ID repeat being allocated the same hexabundle or nearest size bundle and recorded
                    MagnetDict[i.__class__.__name__ + ' ' + str(int(i.index))]['hexabundle'] = case
                    hexabundleRepeatwithsameID.insert(len(hexabundleRepeatwithsameID), case)
                    print(hexabundleRepeatwithsameID)

                    # Respective hexabundle being taken out of hexabundle dictionary as it's allocated already
                    hexabundleDict.pop(case, None)

                    # Index being updated with hexabundle allocation
                    if ('A' <= case <= 'G'):
                        hexabundleIndexAAOmega += 1
                    elif ('I' <= case <= 'T'):
                        hexabundleIndexHector += 1

                # Galaxy ID match not found in the records
                elif galaxyIDrecord[recordID] not in hexabundleRepeatwithsameID:

                    # Galaxy without ID repeat being allocated the same hexabundle and recorded
                    MagnetDict[i.__class__.__name__ + ' ' + str(int(i.index))]['hexabundle'] = galaxyIDrecord[recordID]
                    hexabundleRepeatwithsameID.insert(len(hexabundleRepeatwithsameID),galaxyIDrecord[recordID])
                    print(hexabundleRepeatwithsameID)

                    # Respective hexabundle being taken out of hexabundle dictionary as it's allocated already
                    hexabundleDict.pop(galaxyIDrecord[recordID], None)

                    # Index being updated with hexabundle allocation
                    if ('A' <= galaxyIDrecord[recordID] <= 'G'):
                        hexabundleIndexAAOmega += 1
                    elif ('I' <= galaxyIDrecord[recordID] <= 'T'):
                        hexabundleIndexHector += 1

                    print(hexabundleIndexHector)
                    print(MagnetDict[i.__class__.__name__+' '+str(int(i.index))]['hexabundle'])


        # To avoid multiple recording of galaxy ID, recording only the ones not available in galaxy ID records
        if str(i.IDs) != 'nan' and i.__class__.__name__ == 'rectangular_magnet' and (i.IDs not in galaxyIDrecord):
                galaxyIDrecord[round(i.IDs)] = ''

    #     # Arrays to store the respective parameters for visualizing the hexabundle allocation criteria for convenience
    #     if i.__class__.__name__ == 'rectangular_magnet' and i.galaxyORstar == 1 and (12 > i.Re > 5) and len(answer) < 7 and i.mu_1re < 23:
    #         answer = answer + [i.mu_1re]
    #     elif i.__class__.__name__ == 'rectangular_magnet' and i.galaxyORstar == 1 and i.Mstar < 9.5 and len(answer2) < 12:
    #         answer2 = answer2 + [i.Mstar]
    #     elif i.__class__.__name__ == 'rectangular_magnet' and i.galaxyORstar == 1 and len(answer3) < (21 - len(answer) - len(answer2)):
    #         answer3 = answer3 + [i.mu_1re]
    #
    #     # sorted index of the arrays as per the respective parameters
    #     mu_1re_INDEX = np.argsort(np.argsort(answer))
    #     Mstar_INDEX = np.argsort(np.argsort(answer2))
    #     mu_1re_INDEX_remaining = np.argsort(np.argsort(answer3))
    #
    # #print(answer)
    # print(mu_1re_INDEX)
    # print(Mstar_INDEX)
    # print(mu_1re_INDEX_remaining)
    #print(string.ascii_uppercase[25])
    #print(MagnetDict.keys())


    # allocation of 'I' and 'U' hexabundles to the stars
    MagnetDict, hexabundleDict = star_hexabundle_allocation(MagnetDict, hexabundleDict)

    # getting a count of the probe types- hexa and guides
    GuidesCount, HexaCount = get_probe_count(MagnetDict)

    print(hexabundleDict)
    print(GuidesCount)
    print(HexaCount)

    Mu_1re_Sort = sorted(MagnetDict.items(), key=lambda x: x[1]['mu_1re'])


    ## FIRST CRITERIA FOR HEXABUNDLE ALLOCATION: AAOmega

    Re_Sort = sorted(MagnetDict.items(), key=lambda x: x[1]['Re'], reverse=True)  # largest Re at first
    for magnet in Re_Sort:

        # allocating AAOmega hexabundles to rectangular magnets of Re within range 5-12 and mu_1re less than 23
        if ('rectangular_magnet' in magnet[0]) and (magnet[1]['hexabundle'] == 'NA') and (12 > magnet[1]['Re'] > 5) \
        and hexabundleIndexAAOmega < 7 and magnet[1]['mu_1re'] < 23:

            for hexabundle in hexabundleDict:
                if hexabundleDict[hexabundle][0] == 'A':
                    MagnetDict[magnet[0]]['hexabundle'] = hexabundle
                    print(magnet)
                    hexabundleIndexAAOmega += 1
                    break
            # removing the allocated hexabundle from the hexabundle dictionary
            del hexabundleDict[MagnetDict[magnet[0]]['hexabundle']]

    print(hexabundleDict)
    print(hexabundleIndexAAOmega)

    ## SECOND CRITERIA FOR HEXABUNDLE ALLOCATION: Spector
    Mstar_Sort = sorted(MagnetDict.items(), key=lambda x: x[1]['Mstar'])
    for magnet in Mstar_Sort:

        # allocating Spector hexabundles to rectangular magnets of Mstar less than 9.5
        if ('rectangular_magnet' in magnet[0]) and (magnet[1]['hexabundle'] == 'NA') and (magnet[1]['Mstar'] < 9.5) \
        and hexabundleIndexHector < 12:

            for hexabundle in reversed(hexabundleDict):
                if hexabundleDict[hexabundle][0] == 'S':
                    MagnetDict[magnet[0]]['hexabundle'] = hexabundle
                    print(magnet)
                    hexabundleIndexHector += 1
                    break
            # removing the allocated hexabundle from the hexabundle dictionary
            del hexabundleDict[MagnetDict[magnet[0]]['hexabundle']]

    print(hexabundleDict)
    print(hexabundleIndexHector)

    # THIRD CRITERIA FOR HEXABUNDLE ALLOCATION FOR REMAINING AAOMEGA AND SPECTOR
    hexabundleIndexThirdCriteria = (HexaCount-2) - hexabundleIndexAAOmega - hexabundleIndexHector
    for magnet in Re_Sort:

        # allocating hexabundles larger than 61-cores to rectangular magnets based on larger Re values
        if ('rectangular_magnet' in magnet[0]) and (magnet[1]['hexabundle'] == 'NA') \
                and hexabundleIndexRemaining < hexabundleIndexThirdCriteria:

            for hexabundle in hexabundleDict:
                if hexabundleDict[hexabundle][1] != '61':
                    MagnetDict[magnet[0]]['hexabundle'] = hexabundle
                    print(magnet)
                    hexabundleIndexRemaining += 1
                    break
            # removing the allocated hexabundle from the hexabundle dictionary
            hexabundleDict.pop(MagnetDict[magnet[0]]['hexabundle'], None)#del hexabundleDict[MagnetDict[magnet[0]]['hexabundle']]


    print(hexabundleDict)
    print(hexabundleIndexRemaining)
    print(hexabundleIndexThirdCriteria)
    print('\n')

    i = 0
    j = 0
    for magnet in Mstar_Sort:

        # allocating rest of 61-core hexabundles with lower Mstar value going to Spector
        if ('rectangular_magnet' in magnet[0]) and (magnet[1]['hexabundle'] == 'NA') \
        and hexabundleIndexRemaining < hexabundleIndexThirdCriteria:

            print(sorted(hexabundleDict.items(), key = lambda p: p[0][0],reverse=True))
            for item in sorted(hexabundleDict.items(), key = lambda p: p[0][0],reverse=True):
                print(item[0])
                if hexabundleDict[item[0]][0] == 'S':
                    MagnetDict[magnet[0]]['hexabundle'] = item[0]
                    print(magnet)
                    hexabundleIndexRemaining += 1
                    break
                elif hexabundleDict[item[0]][0] == 'A':
                    MagnetDict[magnet[0]]['hexabundle'] = item[0]
                    print(magnet)
                    hexabundleIndexRemaining += 1
                    break
            # removing the allocated hexabundle from the hexabundle dictionary
            hexabundleDict.pop(MagnetDict[magnet[0]]['hexabundle'], None)  # del hexabundleDict[MagnetDict[magnet[0]]['hexabundle']]

    print(hexabundleDict)

    # I AND D hexabundles of 91 cores being checked and resolved to make hexabundle 'I' have lower Mstar value
    MagnetDict = cores91_hexabundles_sortedByMstar(MagnetDict)

    # checking for repeats of hexabundle
    result = check_for_hexabundle_allocation_repeats(MagnetDict)
    # resolving repeats of hexabundle
    MagnetDict, hexabundleRepeatwithsameID, indexx = resolve_hexabundle_allocation_repeats(result, \
                                                                                hexabundleRepeatwithsameID,MagnetDict)

    # checking for hexabundles not allocated initially
    unallocatedOnes = check_for_unallocated_hexabundle(MagnetDict, HexaCount)
    # resolving hexabundles not allocated initially
    MagnetDict = resolve_unallocated_hexabundle(unallocatedOnes,result,MagnetDict,indexx)


    # recording hexabundle as per galaxy records; IDs in current tile have been recorded above
    galaxyIDrecord, MagnetDict = update_galaxyIDrecord_withHexabundle(galaxyIDrecord,MagnetDict)

    return galaxyIDrecord, MagnetDict

def overall_hexabundle_size_allocation_operation_version3_largerBundlePriority(all_magnets, galaxyIDrecord, clusterNum, tileNum, flagsFile):

    hexabundleIndexAAOmega = 0
    hexabundleIndexHector = 0
    hexabundleIndexRemaining = 0

    answer = []
    answer2 = []

    hexabundleRepeatwithsameID = []

    hexabundleDict = {'A':'A169','B':'A169','C':'A127','D':'A91','E':'A61','F':'A61','G':'A61','H':'A37','I':'S91','J':'S61','K':'S61','L':'S61','M':'S61','N':'S61','O':'S61','P':'S61','Q':'S61','R':'S61','S':'S61','T':'S61','U':'S37'}

    for hexabundleLetter in hexabundleDict:
        hexabundleDict[hexabundleLetter] = re.split('(\d+)',hexabundleDict[hexabundleLetter])[0:2]

    # Creating magnet dictionary with all necessary parameters required for hexabundle allocation
    MagnetDict = create_Magnet_Dictionary(all_magnets)

    for i in all_magnets:

        ## Check for match of galaxies with same ID
        for recordID in galaxyIDrecord:

            # searching each magnet ID in the records
            if i.IDs == recordID and i.__class__.__name__ == 'rectangular_magnet' \
            and galaxyIDrecord[recordID] != 'U' and galaxyIDrecord[recordID] != 'H':
                print('Galaxy ID match found with previous tiles !')
                # print(recordID)
                # print(galaxyIDrecord[recordID])

                # Galaxy ID match found in the records
                if galaxyIDrecord[recordID] in hexabundleRepeatwithsameID:

                    # storing hexabundle letter in the 'case' variable
                    case = galaxyIDrecord[recordID]

                    # Allocating the next largest hexabundle when a repeat of same hexabundle allocation occurred
                    while (case in hexabundleRepeatwithsameID):
                        # AAOmega hexabundle repeat- shifting to next largest same type of bundle
                        if 'A' < case <= 'G':
                            case = chr(ord(case)-1)
                        # Spector hexabundle repeat- shifting to next largest same type of bundle
                        elif 'I' <= case < 'T':
                            case = chr(ord(case)+1)
                        # All AAOmega hexabundles allocated, so shifting to Spector and recording as flag
                        elif hexabundleIndexAAOmega == 7:
                            case = 'J'
                            with open(flagsFile, 'a') as fp:
                                writeToFile = 'ID CHECK. All AAOmega  hexabundles allocated, so shifting to Spector. Cluster/Batch :' + str(clusterNum) + ',tile ' + str(tileNum) + '; ' \
                                              + i.__class__.__name__+' '+str(int(i.index))+', galaxy ID:'+str(recordID) \
                                              + ', hexabundle: ' + str(galaxyIDrecord[recordID]) + '\n'
                                fp.write(writeToFile)
                            print('ALL AAOMEGA HAS BEEN ALLOCATED SO HECTOR WILL BE SELECTED NOW')
                        # All Spector hexabundles allocated, so shifting to AAOmega and recording as flag
                        elif hexabundleIndexHector == 12:
                            case = 'G'
                            with open(flagsFile, 'a') as fp:
                                writeToFile = 'ID CHECK. All Spector hexabundles allocated, so shifting to AAOmega. Cluster/Batch :'+str(clusterNum)+',tile '+str(tileNum)+'; ' \
                                              + i.__class__.__name__+' '+str(int(i.index))+', galaxy ID:'+str(recordID) \
                                              + ', hexabundle: '+ str(galaxyIDrecord[recordID]) + '\n'
                                fp.write(writeToFile)
                            print('ALL SPECTOR HAS BEEN ALLOCATED SO AAOMEGA WILL BE SELECTED NOW')
                        # Move back to smallest AAOmega hexabundle in case of A and B both being in records already
                        elif case == 'A':
                            if ('B' in hexabundleRepeatwithsameID):
                                case = 'G'
                            elif  ('B' not in hexabundleRepeatwithsameID):
                                case = 'B'
                        # Move back to first Spector hexabundle in case of last one being in records already
                        elif case == 'T':
                            case = 'I'
                        else:
                        # Record the exceptional case of an unresolved repeat as a flag
                            with open(flagsFile, 'a') as fp:
                                writeToFile = 'EXCEPTIONAL CASE OF UNRESOLVED REPEAT !!! \n'
                                fp.write(writeToFile)
                            break

                    # Galaxy with ID repeat being allocated the same hexabundle or nearest size bundle and recorded
                    MagnetDict[i.__class__.__name__ + ' ' + str(int(i.index))]['hexabundle'] = case
                    hexabundleRepeatwithsameID.insert(len(hexabundleRepeatwithsameID), case)
                    print(hexabundleRepeatwithsameID)

                    # Respective hexabundle being taken out of hexabundle dictionary as it's allocated already
                    hexabundleDict.pop(case, None)

                    # Index being updated with hexabundle allocation
                    if ('A' <= case <= 'G'):
                        hexabundleIndexAAOmega += 1
                    elif ('I' <= case <= 'T'):
                        hexabundleIndexHector += 1

                # Galaxy ID match not found in the records
                elif galaxyIDrecord[recordID] not in hexabundleRepeatwithsameID:

                    # Galaxy without ID repeat being allocated the same hexabundle and recorded
                    MagnetDict[i.__class__.__name__ + ' ' + str(int(i.index))]['hexabundle'] = galaxyIDrecord[recordID]
                    hexabundleRepeatwithsameID.insert(len(hexabundleRepeatwithsameID),galaxyIDrecord[recordID])
                    print(hexabundleRepeatwithsameID)

                    # Respective hexabundle being taken out of hexabundle dictionary as it's allocated already
                    hexabundleDict.pop(galaxyIDrecord[recordID], None)

                    # Index being updated with hexabundle allocation
                    if ('A' <= galaxyIDrecord[recordID] <= 'G'):
                        hexabundleIndexAAOmega += 1
                    elif ('I' <= galaxyIDrecord[recordID] <= 'T'):
                        hexabundleIndexHector += 1

                    print(hexabundleIndexHector)
                    print(MagnetDict[i.__class__.__name__+' '+str(int(i.index))]['hexabundle'])


        # To avoid multiple recording of galaxy ID, recording only the ones not available in galaxy ID records
        if str(i.IDs) != 'nan' and i.__class__.__name__ == 'rectangular_magnet' and (i.IDs not in galaxyIDrecord):
                galaxyIDrecord[round(i.IDs)] = ''

        # Getting an array count of mu_1re graeter and less than 22
        if i.__class__.__name__ == 'rectangular_magnet' and i.galaxyORstar == 1 and i.Re < 12:
            answer = answer + [i.Re]
        elif i.__class__.__name__ == 'rectangular_magnet' and i.galaxyORstar == 1 and i.Re >= 12:
            answer2 = answer2 + [i.Re]

        Re_lt12_INDEX = np.argsort(np.argsort(answer))
        Re_gt12_INDEX = np.argsort(np.argsort(answer2))


    print("\nRe below and above 12:  ")
    print(Re_lt12_INDEX)
    print(Re_gt12_INDEX)

    # allocation of 'I' and 'U' hexabundles to the stars
    MagnetDict, hexabundleDict = star_hexabundle_allocation(MagnetDict, hexabundleDict)

    # getting a count of the probe types- hexa and guides
    GuidesCount, HexaCount = get_probe_count(MagnetDict)

    print(hexabundleDict)
    print(GuidesCount)
    print(HexaCount)

    Re_Sort = sorted(MagnetDict.items(), key=lambda x: x[1]['Re'], reverse=True)  # largest Re at first
    Mstar_Sort = sorted(MagnetDict.items(), key=lambda x: x[1]['Mstar'])

    # Getting a count of hexabundles larger than 61
    large_bundle_count = 0
    for hexabundle in hexabundleDict:
        if int(hexabundleDict[hexabundle][1]) > 61:
            large_bundle_count += 1

    print(large_bundle_count)
    large_bundle_index = 0
    # Allocating all hexabundles greater than 61 based on Re<12
    for magnet in Re_Sort:
        if ('rectangular_magnet' in magnet[0]) and (magnet[1]['galaxyOrStar'] == 1) and (
                magnet[1]['hexabundle'] == 'NA') and (magnet[1]['Re'] < 12) and large_bundle_index < large_bundle_count:
            for hexabundle in hexabundleDict:
                if int(hexabundleDict[hexabundle][1]) > 61:  # check for larger hexabundles
                    MagnetDict[magnet[0]]['hexabundle'] = hexabundle
                    print(magnet)
                    print('Larger hexabundles')
                    large_bundle_index += 1
                    if hexabundleDict[hexabundle][0] == 'A':
                        hexabundleIndexAAOmega += 1
                    elif hexabundleDict[hexabundle][0] == 'S':
                        hexabundleIndexHector += 1
                    break
            del hexabundleDict[MagnetDict[magnet[0]]['hexabundle']]

    # In case any larger hexabundles more than 61-cores have not been allocated yet
    # (which is highly unlikely and might happen due to repeats)
    for magnet in Re_Sort:
        if ('rectangular_magnet' in magnet[0]) and (magnet[1]['galaxyOrStar'] == 1) and (
                magnet[1]['hexabundle'] == 'NA') and large_bundle_index < large_bundle_count:
            for hexabundle in hexabundleDict:
                if int(hexabundleDict[hexabundle][1]) > 61:  # check for larger hexabundles
                    MagnetDict[magnet[0]]['hexabundle'] = hexabundle
                    print(magnet)
                    print('Larger hexabundles but with more than Re=12')
                    large_bundle_index += 1
                    if hexabundleDict[hexabundle][0] == 'A':
                        hexabundleIndexAAOmega += 1
                    elif hexabundleDict[hexabundle][0] == 'S':
                        hexabundleIndexHector += 1
                    with open(flagsFile, 'a') as fp:
                        writeToFile = 'RARE case of larger hexabundles greater than 61-cores left for allocation after Re<12: ' \
                                      'Cluster/Batch :' + str(clusterNum) + ',tile ' + str(tileNum) + '; ' \
                                      + str(magnet[0]) + ', ID:' + str(recordID) + ', hexabundle: ' + str(
                            hexabundle) + '\n'
                        fp.write(writeToFile)
                    break
            del hexabundleDict[MagnetDict[magnet[0]]['hexabundle']]

    # I AND D hexabundles of 91 cores being checked and resolved to make hexabundle 'I' have lower Mstar value
    MagnetDict = cores91_hexabundles_sortedByMstar(MagnetDict)

    print('\nLarger hexabundles more than 61 have been allocated.\n')

    X = 22.5
    print('mu_1re cutoff, X = ' + str(X))
    # Allocating all 61 hexabundles based on mu_1Re < X
    for magnet in Mstar_Sort:  # Spector
        if ('rectangular_magnet' in magnet[0]) and (magnet[1]['galaxyOrStar'] == 1) and (
                magnet[1]['hexabundle'] == 'NA') and (magnet[1]['mu_1re'] < X) and hexabundleIndexHector < 12:
            for hexabundle in hexabundleDict:
                if int(hexabundleDict[hexabundle][1]) == 61 and hexabundleDict[hexabundle][0] == 'S':  # check for larger hexabundles
                    MagnetDict[magnet[0]]['hexabundle'] = hexabundle
                    print(magnet)
                    print('mu_1re<X: 61 hexabundles to Spector')
                    hexabundleIndexHector += 1
                    break
            del hexabundleDict[MagnetDict[magnet[0]]['hexabundle']]

    for magnet in Mstar_Sort:  # AAOmega
        if ('rectangular_magnet' in magnet[0]) and (magnet[1]['galaxyOrStar'] == 1) and (
                magnet[1]['hexabundle'] == 'NA') and (magnet[1]['mu_1re'] < X) and hexabundleIndexAAOmega < 7:
            for hexabundle in hexabundleDict:
                if int(hexabundleDict[hexabundle][1]) == 61 and hexabundleDict[hexabundle][0] == 'A':  # check for larger hexabundles
                    MagnetDict[magnet[0]]['hexabundle'] = hexabundle
                    print(magnet)
                    print('mu_1re<X: 61 hexabundles to AAOmega')
                    hexabundleIndexAAOmega += 1
                    break
            del hexabundleDict[MagnetDict[magnet[0]]['hexabundle']]

    print('\n61-core hexabundles based with mu_1Re<X have been allocated \n')

    # Allocating rest of 61 hexabundles based on mu_1Re > X
    for magnet in Mstar_Sort:  # Spector
        if ('rectangular_magnet' in magnet[0]) and (magnet[1]['galaxyOrStar'] == 1) and (
                magnet[1]['hexabundle'] == 'NA') and (magnet[1]['mu_1re'] >= X) and hexabundleIndexHector < 12:
            for hexabundle in hexabundleDict:
                if int(hexabundleDict[hexabundle][1]) == 61 and hexabundleDict[hexabundle][0] == 'S':  # check for larger hexabundles
                    MagnetDict[magnet[0]]['hexabundle'] = hexabundle
                    print(magnet)
                    print('mu_1re>X: 61 hexabundles to Spector')
                    hexabundleIndexHector += 1
                    break
            del hexabundleDict[MagnetDict[magnet[0]]['hexabundle']]

    for magnet in Mstar_Sort:  # AAOmega
        if ('rectangular_magnet' in magnet[0]) and (magnet[1]['galaxyOrStar'] == 1) and (
                magnet[1]['hexabundle'] == 'NA') and (magnet[1]['mu_1re'] >= X) and hexabundleIndexAAOmega < 7:
            for hexabundle in hexabundleDict:
                if int(hexabundleDict[hexabundle][1]) == 61 and hexabundleDict[hexabundle][0] == 'A':  # check for larger hexabundles
                    MagnetDict[magnet[0]]['hexabundle'] = hexabundle
                    print(magnet)
                    print('mu_1re>X: 61 hexabundles to AAOmega')
                    hexabundleIndexAAOmega += 1
                    break
            del hexabundleDict[MagnetDict[magnet[0]]['hexabundle']]

    print('\nAll 61 hexabundles have been allocated as well.\n')
    print(hexabundleDict)

    # checking for repeats of hexabundle
    result = check_for_hexabundle_allocation_repeats(MagnetDict)
    # resolving repeats of hexabundle
    MagnetDict, hexabundleRepeatwithsameID, indexx = resolve_hexabundle_allocation_repeats(result, \
                                                                                hexabundleRepeatwithsameID,MagnetDict)

    # checking for hexabundles not allocated initially
    unallocatedOnes = check_for_unallocated_hexabundle(MagnetDict, HexaCount)
    # resolving hexabundles not allocated initially
    MagnetDict = resolve_unallocated_hexabundle(unallocatedOnes,result,MagnetDict,indexx)


    # recording hexabundle as per galaxy records; IDs in current tile have been recorded above
    galaxyIDrecord, MagnetDict = update_galaxyIDrecord_withHexabundle(galaxyIDrecord,MagnetDict)

    return galaxyIDrecord, MagnetDict

def overall_hexabundle_size_allocation_operation_version2_median(all_magnets, galaxyIDrecord, clusterNum, tileNum, flagsFile):

    hexabundleIndexAAOmega = 0
    hexabundleIndexHector = 0
    hexabundleIndexRemaining = 0

    answer = []
    answer2 = []
    answer3 = []

    hexabundleRepeatwithsameID = []

    hexabundleDict = {'A':'A169','B':'A169','C':'A127','D':'A91','E':'A61','F':'A61','G':'A61','H':'A37','I':'S91','J':'S61','K':'S61','L':'S61','M':'S61','N':'S61','O':'S61','P':'S61','Q':'S61','R':'S61','S':'S61','T':'S61','U':'S37'}

    for hexabundleLetter in hexabundleDict:
        hexabundleDict[hexabundleLetter] = re.split('(\d+)',hexabundleDict[hexabundleLetter])[0:2]

    # Creating magnet dictionary with all necessary parameters required for hexabundle allocation
    MagnetDict = create_Magnet_Dictionary(all_magnets)

    for i in all_magnets:

        ## Check for match of galaxies with same ID
        for recordID in galaxyIDrecord:

            # searching each magnet ID in the records
            if i.IDs == recordID and i.__class__.__name__ == 'rectangular_magnet' \
            and galaxyIDrecord[recordID] != 'U' and galaxyIDrecord[recordID] != 'H':
                print('Galaxy ID match found with previous tiles !')
                # print(recordID)
                # print(galaxyIDrecord[recordID])

                # Galaxy ID match found in the records
                if galaxyIDrecord[recordID] in hexabundleRepeatwithsameID:

                    # storing hexabundle letter in the 'case' variable
                    case = galaxyIDrecord[recordID]

                    # Allocating the next largest hexabundle when a repeat of same hexabundle allocation occurred
                    while (case in hexabundleRepeatwithsameID):
                        # AAOmega hexabundle repeat- shifting to next largest same type of bundle
                        if 'A' < case <= 'G':
                            case = chr(ord(case)-1)
                        # Spector hexabundle repeat- shifting to next largest same type of bundle
                        elif 'I' <= case < 'T':
                            case = chr(ord(case)+1)
                        # All AAOmega hexabundles allocated, so shifting to Spector and recording as flag
                        elif hexabundleIndexAAOmega == 7:
                            case = 'J'
                            with open(flagsFile, 'a') as fp:
                                writeToFile = 'ID CHECK. All AAOmega  hexabundles allocated, so shifting to Spector. Cluster/Batch :' + str(clusterNum) + ',tile ' + str(tileNum) + '; ' \
                                              + i.__class__.__name__+' '+str(int(i.index))+', galaxy ID:'+str(recordID) \
                                              + ', hexabundle: ' + str(galaxyIDrecord[recordID]) + '\n'
                                fp.write(writeToFile)
                            print('ALL AAOMEGA HAS BEEN ALLOCATED SO HECTOR WILL BE SELECTED NOW')
                        # All Spector hexabundles allocated, so shifting to AAOmega and recording as flag
                        elif hexabundleIndexHector == 12:
                            case = 'G'
                            with open(flagsFile, 'a') as fp:
                                writeToFile = 'ID CHECK. All Spector hexabundles allocated, so shifting to AAOmega. Cluster/Batch :'+str(clusterNum)+',tile '+str(tileNum)+'; ' \
                                              + i.__class__.__name__+' '+str(int(i.index))+', galaxy ID:'+str(recordID) \
                                              + ', hexabundle: '+ str(galaxyIDrecord[recordID]) + '\n'
                                fp.write(writeToFile)
                            print('ALL SPECTOR HAS BEEN ALLOCATED SO AAOMEGA WILL BE SELECTED NOW')
                        # Move back to smallest AAOmega hexabundle in case of A and B both being in records already
                        elif case == 'A':
                            if ('B' in hexabundleRepeatwithsameID):
                                case = 'G'
                            elif  ('B' not in hexabundleRepeatwithsameID):
                                case = 'B'
                        # Move back to first Spector hexabundle in case of last one being in records already
                        elif case == 'T':
                            case = 'I'
                        else:
                        # Record the exceptional case of an unresolved repeat as a flag
                            with open(flagsFile, 'a') as fp:
                                writeToFile = 'EXCEPTIONAL CASE OF UNRESOLVED REPEAT !!! \n'
                                fp.write(writeToFile)
                            break

                    # Galaxy with ID repeat being allocated the same hexabundle or nearest size bundle and recorded
                    MagnetDict[i.__class__.__name__ + ' ' + str(int(i.index))]['hexabundle'] = case
                    hexabundleRepeatwithsameID.insert(len(hexabundleRepeatwithsameID), case)
                    print(hexabundleRepeatwithsameID)

                    # Respective hexabundle being taken out of hexabundle dictionary as it's allocated already
                    hexabundleDict.pop(case, None)

                    # Index being updated with hexabundle allocation
                    if ('A' <= case <= 'G'):
                        hexabundleIndexAAOmega += 1
                    elif ('I' <= case <= 'T'):
                        hexabundleIndexHector += 1

                # Galaxy ID match not found in the records
                elif galaxyIDrecord[recordID] not in hexabundleRepeatwithsameID:

                    # Galaxy without ID repeat being allocated the same hexabundle and recorded
                    MagnetDict[i.__class__.__name__ + ' ' + str(int(i.index))]['hexabundle'] = galaxyIDrecord[recordID]
                    hexabundleRepeatwithsameID.insert(len(hexabundleRepeatwithsameID),galaxyIDrecord[recordID])
                    print(hexabundleRepeatwithsameID)

                    # Respective hexabundle being taken out of hexabundle dictionary as it's allocated already
                    hexabundleDict.pop(galaxyIDrecord[recordID], None)

                    # Index being updated with hexabundle allocation
                    if ('A' <= galaxyIDrecord[recordID] <= 'G'):
                        hexabundleIndexAAOmega += 1
                    elif ('I' <= galaxyIDrecord[recordID] <= 'T'):
                        hexabundleIndexHector += 1

                    print(hexabundleIndexHector)
                    print(MagnetDict[i.__class__.__name__+' '+str(int(i.index))]['hexabundle'])


        # To avoid multiple recording of galaxy ID, recording only the ones not available in galaxy ID records
        if str(i.IDs) != 'nan' and i.__class__.__name__ == 'rectangular_magnet' and (i.IDs not in galaxyIDrecord):
                galaxyIDrecord[round(i.IDs)] = ''

        # Getting an array count of mu_1re greater and less than 22
        if i.__class__.__name__ == 'rectangular_magnet' and i.galaxyORstar == 1 and i.mu_1re < 22:
            answer = answer + [i.mu_1re]
        elif i.__class__.__name__ == 'rectangular_magnet' and i.galaxyORstar == 1 and i.mu_1re >= 22:
            answer2 = answer2 + [i.mu_1re]

        mu_1re_lt22_INDEX = np.argsort(np.argsort(answer))
        mu_1re_gt22_INDEX = np.argsort(np.argsort(answer2))

    # print(answer)
    print("\nmu_1re above and below 22:  ")
    print(mu_1re_gt22_INDEX)
    print(mu_1re_lt22_INDEX)

    # allocation of 'I' and 'U' hexabundles to the stars
    MagnetDict, hexabundleDict = star_hexabundle_allocation(MagnetDict, hexabundleDict)

    # getting a count of the probe types- hexa and guides
    GuidesCount, HexaCount = get_probe_count(MagnetDict)

    print(hexabundleDict)
    print(GuidesCount)
    print(HexaCount)

    Re_Sort = sorted(MagnetDict.items(), key=lambda x: x[1]['Re'], reverse=True)  # largest Re at first
    Mstar_Sort = sorted(MagnetDict.items(), key=lambda x: x[1]['Mstar'])

    lowMstar_spector_number = round((12 / 19) * len(mu_1re_lt22_INDEX))
    print(lowMstar_spector_number)
    Mstar_median_lowerhalf = {}
    Mstar_median_higherhalf = {}
    Mstar_Sort_count = 0
    for magnet in Mstar_Sort:
        if ('rectangular_magnet' in magnet[0]) and (magnet[1]['mu_1re'] < 22) and (magnet[1]['galaxyOrStar'] == 1) and (
                Mstar_Sort_count < lowMstar_spector_number):
            key = str(magnet[0])
            Mstar_median_lowerhalf[key] = MagnetDict[key]
            Mstar_Sort_count += 1
        elif ('rectangular_magnet' in magnet[0]) and (magnet[1]['mu_1re'] < 22) and (
                magnet[1]['galaxyOrStar'] == 1) and ((HexaCount - 2) > Mstar_Sort_count >= lowMstar_spector_number):
            key = str(magnet[0])
            Mstar_median_higherhalf[key] = MagnetDict[key]
            Mstar_Sort_count += 1

    print(Mstar_median_lowerhalf)
    print(Mstar_median_higherhalf)
    print(len(Mstar_median_lowerhalf))
    print(len(Mstar_median_higherhalf))
    print('\nMedian based sorting')
    print(lowMstar_spector_number)

    lowMstar_Sort_count = 0
    highMstar_Sort_count = 0

    # Lower mass galaxies with mu_1Re < 22 go to Spector
    for magnet in Re_Sort:
        if ('rectangular_magnet' in magnet[0]) and (magnet[1]['hexabundle'] == 'NA') and (magnet[1]['mu_1re'] < 22):
            for magnet_2nd in Mstar_median_lowerhalf:
                if str(magnet_2nd) == str(magnet[0]) and hexabundleIndexHector < 12:
                    print('MATCH')
                    for hexabundle in hexabundleDict:
                        if hexabundleDict[hexabundle][0] == 'S':  # check for first Spector
                            MagnetDict[magnet[0]]['hexabundle'] = hexabundle
                            print(magnet)
                            hexabundleIndexHector += 1
                            lowMstar_Sort_count += 1
                            break
                    del hexabundleDict[MagnetDict[magnet[0]]['hexabundle']]

    # Higher mass with mu_1Re < 22 go to AAOmega
    for magnet in Re_Sort:
        if ('rectangular_magnet' in magnet[0]) and (magnet[1]['hexabundle'] == 'NA') and (magnet[1]['mu_1re'] < 22):
            for magnet_2nd in Mstar_median_higherhalf:
                if str(magnet_2nd) == str(magnet[0]) and hexabundleIndexAAOmega < 7:
                    print('MATCH2')
                    for hexabundle in hexabundleDict:
                        if hexabundleDict[hexabundle][0] == 'A':  # check for first AAOmega
                            MagnetDict[magnet[0]]['hexabundle'] = hexabundle
                            print(magnet)
                            hexabundleIndexAAOmega += 1
                            highMstar_Sort_count += 1
                            break
                    del hexabundleDict[MagnetDict[magnet[0]]['hexabundle']]

    # Remaining galaxies with mu_1re<22 placed on remaining bundles based on largest â€œRe" placed on the largest bundles
    for magnet in Re_Sort:
        if ('rectangular_magnet' in magnet[0]) and (magnet[1]['hexabundle'] == 'NA') and magnet[1]['mu_1re'] < 22:
            for hexabundle in hexabundleDict:
                if int(hexabundleDict[hexabundle][1]) > 61:  # check for greater hexabundle
                    MagnetDict[magnet[0]]['hexabundle'] = hexabundle
                    print(magnet)
                    if hexabundleDict[hexabundle][0] == 'A':
                        hexabundleIndexAAOmega += 1
                    elif hexabundleDict[hexabundle][0] == 'S':
                        hexabundleIndexHector += 1
                    # highMstar_Sort_count += 1
                    break
                elif int(hexabundleDict[hexabundle][1]) == 61:
                    MagnetDict[magnet[0]]['hexabundle'] = hexabundle
                    print(magnet)
                    if hexabundleDict[hexabundle][0] == 'A':
                        hexabundleIndexAAOmega += 1
                    elif hexabundleDict[hexabundle][0] == 'S':
                        hexabundleIndexHector += 1
                    # highMstar_Sort_count += 1
                    break
            del hexabundleDict[MagnetDict[magnet[0]]['hexabundle']]

    print(hexabundleIndexAAOmega)
    print(hexabundleIndexHector)

    print(lowMstar_Sort_count)
    print(highMstar_Sort_count)

    print('Galaxies with less than Mu_1re=22 allocated\n')

    # Remaining galaxies with mu_1re>22 placed on remaining bundles based on largest â€œRe" placed on the largest bundles
    for magnet in Re_Sort:
        if ('rectangular_magnet' in magnet[0]) and (magnet[1]['hexabundle'] == 'NA') and magnet[1]['mu_1re'] >= 22:
            for hexabundle in hexabundleDict:
                if int(hexabundleDict[hexabundle][1]) > 61:  # check for greater hexabundle
                    MagnetDict[magnet[0]]['hexabundle'] = hexabundle
                    print(magnet)
                    if hexabundleDict[hexabundle][0] == 'A':
                        hexabundleIndexAAOmega += 1
                    elif hexabundleDict[hexabundle][0] == 'S':
                        hexabundleIndexHector += 1
                    # highMstar_Sort_count += 1
                    break
                elif int(hexabundleDict[hexabundle][1]) == 61:
                    MagnetDict[magnet[0]]['hexabundle'] = hexabundle
                    print(magnet)
                    if hexabundleDict[hexabundle][0] == 'A':
                        hexabundleIndexAAOmega += 1
                    elif hexabundleDict[hexabundle][0] == 'S':
                        hexabundleIndexHector += 1
                    # highMstar_Sort_count += 1
                    break
            del hexabundleDict[MagnetDict[magnet[0]]['hexabundle']]

    print('Galaxies with more than Mu_1re=22 allocated\n')

    print(hexabundleDict)

    # I AND D hexabundles of 91 cores being checked and resolved to make hexabundle 'I' have lower Mstar value
    MagnetDict = cores91_hexabundles_sortedByMstar(MagnetDict)

    # checking for repeats of hexabundle
    result = check_for_hexabundle_allocation_repeats(MagnetDict)
    # resolving repeats of hexabundle
    MagnetDict, hexabundleRepeatwithsameID, indexx = resolve_hexabundle_allocation_repeats(result, \
                                                                                hexabundleRepeatwithsameID,MagnetDict)

    # checking for hexabundles not allocated initially
    unallocatedOnes = check_for_unallocated_hexabundle(MagnetDict, HexaCount)
    # resolving hexabundles not allocated initially
    MagnetDict = resolve_unallocated_hexabundle(unallocatedOnes,result,MagnetDict,indexx)


    # recording hexabundle as per galaxy records; IDs in current tile have been recorded above
    galaxyIDrecord, MagnetDict = update_galaxyIDrecord_withHexabundle(galaxyIDrecord,MagnetDict)

    return galaxyIDrecord, MagnetDict


def create_Magnet_Dictionary(all_magnets):

    MagnetDict = {}

    # storing all respective parameters to create a magnet dictionary
    for i in all_magnets:
        MagnetDict[i.__class__.__name__+' '+str(int(i.index))] = {
                    'index': int(i.index),
                    'galaxyOrStar': i.galaxyORstar,
                    'Re': i.Re,
                    'mu_1re': i.mu_1re,
                    'Mstar': i.Mstar,
                    'hexabundle': 'NA',
                    'IDs': i.IDs
                }
    return MagnetDict

def get_probe_count(MagnetDict):

    GuidesCount = 0
    HexaCount = 0

    # getting count of probe type- hexa or guide- based on galaxyorStar parameter
    for magnet in MagnetDict:
        if ('rectangular_magnet' in magnet):
            if (MagnetDict[magnet]['galaxyOrStar'] == 1 or MagnetDict[magnet]['galaxyOrStar'] == 0):
                HexaCount += 1
            else:
                GuidesCount += 1

    return GuidesCount, HexaCount

def star_hexabundle_allocation(MagnetDict,hexabundleDict):

    # identifying stars and allocating the 37-core bundles H and U
    for magnet in MagnetDict:
        if ('rectangular_magnet' in magnet) and (MagnetDict[magnet]['galaxyOrStar'] == 0) and (MagnetDict[magnet]['hexabundle'] == 'NA'):
            for hexabundle in hexabundleDict:
                if hexabundleDict[hexabundle][1] == '37':
                    MagnetDict[magnet]['hexabundle'] = hexabundle
                    break
            del hexabundleDict[MagnetDict[magnet]['hexabundle']]

    return MagnetDict,hexabundleDict

## I AND D hexabundles of 91 cores being checked and resolved to make hexabundle I have lower Mstar value
def cores91_hexabundles_sortedByMstar(MagnetDict):

    countD = 0
    countI = 0
    I_mstar_91Core = 0
    D_mstar_91Core = 0

    for magnet in MagnetDict:
        # print(str(MagnetDict[magnet]['index'])+' '+str(MagnetDict[magnet]['hexabundle']))

        # storing Mstar values and magnet labels of I and D hexabundles of 91 cores
        if MagnetDict[magnet]['hexabundle'] == 'I':
            I_mstar_91Core = MagnetDict[magnet]['Mstar']
            magnet_I = magnet
            countI == 1
        elif MagnetDict[magnet]['hexabundle'] == 'D':
            D_mstar_91core = MagnetDict[magnet]['Mstar']
            magnet_D = magnet
            countD == 1

    # Ensuring I and D hexabundles of 91 cores have been allocated based on Mstar value, with lower one going to I
    if countI == 1 and countD == 1 and I_mstar_91Core > D_mstar_91Core:
        MagnetDict[magnet_I]['hexabundle'] = 'D'
        MagnetDict[magnet_D]['hexabundle'] = 'I'
        print('I AND D SWAPPED')

    return MagnetDict

def resolve_hexabundle_allocation_repeats(result,hexabundleRepeatwithsameID,MagnetDict):

    indexx = 0
    if result != False:
        print('Allocation WRONG WRONG WRONG WRONG WRONG!')
        hexabundleRepeatwithsameID = sorted(hexabundleRepeatwithsameID)
        hexabundleRepeatwithsameID.insert(len(hexabundleRepeatwithsameID), '')
        print(hexabundleRepeatwithsameID)
        sourcelist = 'ABCDEFGIJKLMNOPQRST'
        for repeatOnes in sourcelist:
            checkForRepeat = 0
            for magnet in MagnetDict:
                if MagnetDict[magnet]['hexabundle'] == repeatOnes:
                    if checkForRepeat > 0:
                        MagnetDict[magnet]['hexabundle'] = result[indexx]
                        indexx += 1
                        print('HERE')
                        print(MagnetDict[magnet]['hexabundle'])
                    checkForRepeat += 1
            # print(MagnetDict[magnet]['hexabundle'])
        if checkForRepeat == 0 and indexx == 0:
            sys.exit('WRONG ALLOCATION FOUND!')
    else:
        print('No repeats in allocation.')

    return MagnetDict, hexabundleRepeatwithsameID,indexx

def resolve_unallocated_hexabundle(unallocatedOnes,result,MagnetDict,indexx):

    if unallocatedOnes != False:
        print('NOT ALL ARE ALLOCATED')
        MagnetDict[unallocatedOnes]['hexabundle'] = result[indexx]
        indexx += 1
        print(str(MagnetDict[unallocatedOnes]['hexabundle'])+'has been allocated as it was found unallocated.')
    else:
        print('All Hexabundles are  ALLOCATED!')

    return MagnetDict


def check_for_hexabundle_allocation_repeats(nested_dictionary):
    setOfElems = set()
    repeatLetter = 0
    print(setOfElems)
    for key, value in nested_dictionary.items():

        if type(value) is dict:

            for key1, value1 in value.items():
                if key1 == 'hexabundle':
                    if value1 in setOfElems:
                        repeatLetter = value1
                        print(repeatLetter)
                    elif value1 != 'NA':
                        setOfElems.add(value1)

    sourcelist = 'ABCDEFGIJKLMNOPQRST'
    missingLetter = [letter for letter in sourcelist if letter not in setOfElems]

    if missingLetter != [] and repeatLetter != 0:
        print(missingLetter)
        return missingLetter

    return False

def check_for_unallocated_hexabundle(nested_dictionary, HexaCount):
    setOfElems = set()
    print(setOfElems)
    for key, value in nested_dictionary.items():

        if type(value) is dict:

            for key1, value1 in value.items():
                if (key1 == 'hexabundle'):
                    if (value1 == 'NA') and ('rectangular_magnet' in key) and \
                    (nested_dictionary[key]['galaxyOrStar'] == 1 or nested_dictionary[key]['galaxyOrStar'] == 0):
                    #(0 < nested_dictionary[key]['index'] <= HexaCount):
                        return key

    return False

def update_galaxyIDrecord_withHexabundle(galaxyIDrecord, MagnetDict):

    for recordID in galaxyIDrecord:
        for magnet in MagnetDict:
            if recordID == MagnetDict[magnet]['IDs']:
                galaxyIDrecord[recordID] = MagnetDict[magnet]['hexabundle']

    return galaxyIDrecord, MagnetDict

def hexabundle_allocation_fromMagnetDict(MagnetDict, magnet, guideIndex):
    # copying allocated hexabundle from Magnet Dictionary
    magnet.hexabundle = MagnetDict[magnet.__class__.__name__ + ' ' + str(int(magnet.index))]['hexabundle']

    # allocating hexabundle to guide probes in order GS1,2..
    if (magnet.hexabundle == 'NA') and (magnet.__class__.__name__ == 'rectangular_magnet'):
        magnet.hexabundle = 'GS' + str(guideIndex)
        guideIndex += 1

    return magnet, guideIndex





import numpy as np
from ..hector.probe import probe
from ..problem_operations. offsets import radialPositionOffset


def parse_col(s):
    try:
        return float(s)
    except ValueError:
        return 0

def create_list_of_probes_from_file(file):

    probe_number, \
    IDs, \
    circular_magnet_center_x, \
    circular_magnet_center_y, \
    rads, \
    angs, \
    azAngs, \
    rectangle_magnet_input_orientation, \
    galaxyORstar, \
    Re, \
    mu_1re, \
    Mstar = np.loadtxt(file, skiprows=1, unpack=True, converters={12:parse_col,13:parse_col,24:parse_col}, usecols =[1,9,3,4,5,6,7,8,43,12,24,13])

    list_of_probes = []

    # creating rectangular magnet label
    magnet_label = [None]*len(probe_number)

    # creating hexabundle array
    hexabundle= 'NA' * len(probe_number)

    # creating rotation for pickup and putdown
    rotation_pickup = [None]*len(probe_number)
    rotation_putdown =[0]*len(probe_number)


    i = 0
    for each_probe in probe_number:
        list_of_probes.append(probe(probe_number[i],
                                    [circular_magnet_center_x[i],circular_magnet_center_y[i]],
                                     rectangle_magnet_input_orientation[i],
                                     galaxyORstar[i],
                                     Re[i],
                                     mu_1re[i],
                                     Mstar[i],
                                     magnet_label[i],
                                     hexabundle[i],
                                     rads[i],
                                     rotation_pickup[i],
                                     rotation_putdown[i],
                                     azAngs[i],
                                     IDs[i]))

        i += 1

    return list_of_probes

def create_list_of_circular_and_rectangular_magnets_from_file(file,magnetPair_offset):

    list_of_probes = create_list_of_probes_from_file(file)

    list_of_probes = radialPositionOffset(list_of_probes, magnetPair_offset)
    ## ***** offset adjustments for magnet pair
    # for item in magnetPair_offset:
    #     for each_probe in list_of_probes:
    #         if item[0] == each_probe.index:
    #             # print(each_probe.circular_magnet_center)
    #             # each_probe.circular_magnet_center[0] += item[1]
    #             # each_probe.circular_magnet_center[1] += item[2]
    #             # print(each_probe.circular_magnet_center)
    #
    #             theta = atan(each_probe.circular_magnet_center[1] / each_probe.circular_magnet_center[0])
    #             each_probe.circular_magnet_center = (each_probe.circular_magnet_center[0] + (cos(theta) * item[1]), \
    #                                                  each_probe.circular_magnet_center[1] + (sin(theta) * item[1]))
    ## *****

    list_of_circular_magnet = []

    for each_probe in list_of_probes:
        list_of_circular_magnet.append(each_probe.extract_circular_magnet_parameters())

    list_of_rectangular_magnet = []

    for each_probe in list_of_probes:
        list_of_rectangular_magnet.append(each_probe.extract_rectangular_magnet_parameters())

    return list_of_circular_magnet,list_of_rectangular_magnet

def create_list_of_all_magnets_from_file(file,magnetPair_offset):

    [circular_magnets, rectangular_magnets] = create_list_of_circular_and_rectangular_magnets_from_file(file,magnetPair_offset)

    # # ***** working function for adjusting magnet pair position, needs to be moved to offsets and be controlled from main
    # for magnet in circular_magnets:
    #     if magnet.index == 13:
    #         magnet.center[0] += 20
    #
    # for magnet in rectangular_magnets:
    #     if magnet.index == 13:
    #         magnet.center[0] += 20

    return np.concatenate([circular_magnets, rectangular_magnets])


def get_file(filename):
    return open(filename, "r")


import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import csv
import time


def arrange_guidesFile(fileNameGuides,proxyGuideFile):

    ExtraCols = 1
    # f = open('galaxy_fields/newfile.txt', "w+")
    # f.close()
    df = pd.read_csv(fileNameGuides)
    for i in range(35):
        df['ColFillUp%d' % ExtraCols] = 'NaN'
        ExtraCols += 1

    df.to_csv(proxyGuideFile, index=False, sep=' ', quoting=csv.QUOTE_NONE, escapechar=' ')

    rows = [line.split(' ') for line in open(proxyGuideFile)]
    cols = list(zip(*rows))
    cols.insert(1, ['"1stCol"', 'NaN', 'NaN', 'NaN', 'NaN', 'NaN', 'NaN'])
    cols.insert(1, ['21', '22', '23', '24', '25', '26', '27'])
    cols.insert(1, ['"IDs"', 'NaN', 'NaN', 'NaN', 'NaN', 'NaN', 'NaN'])
    cols = cols[1:]
    rows = list(zip(*cols))

    with open(proxyGuideFile, 'w') as f:
        f.writelines([' '.join(row) for row in rows[1:]])

def merge_hexaAndGuides(fileNameHexa, proxyGuideFile, plate_file):

    with open(fileNameHexa) as fp:
        file1 = fp.read()
    with open(proxyGuideFile) as fp:
        file2 = fp.read()
    file1 += file2

    # with open('GAMA_'+batch+'/Output/Hexa_and_Guides_GAMA_'+batch+'_tile_%03d.txt' % (tileNum),'w+') as fp:
    with open(plate_file, 'w+') as fp:
        fp.write(file1)

    print("Filename being checked for conflicts:")
    print(fileNameHexa)

def create_robotFileArray(positioning_array,robotFile,newrow,fully_blocked_magnets_dictionary):

    positioning_array[:, 8] = [i if i != 'nan' else -999999 for i in positioning_array[:, 8]]
    for i in range(len(positioning_array[:, 8])):
        positioning_array[:, 8][i] = round(float(positioning_array[:, 8][i]))

    robotFilearray = sorted(positioning_array, key=lambda x: x[6])
    robotFilearray = np.insert(robotFilearray, 0, np.array(newrow), 0)

    # rePosition_col = ['0'] * len(positioning_array[:, 8])
    # Creates a list containing 5 lists, each of 8 items, all set to 0
    w, h = len(positioning_array[:, 8])+1,1
    rePosition_col = [['[0]' for x in range(w)] for y in range(h)]

    # rePosition_col = np.transpose(rePosition_col)
    print(len(positioning_array[:, 8]))
    print(len(rePosition_col))
    # positioning_array = np.hstack((positioning_array[:, :8], rePosition_col, positioning_array[:, 8:]))

    for each_magnet in fully_blocked_magnets_dictionary:
        for i in range(len(robotFilearray)):
            if (robotFilearray[i][0] + ' ' + str(robotFilearray[i][9])) == each_magnet:
                rePosition_col[0][i] = '['+str(fully_blocked_magnets_dictionary[each_magnet])+']'

    rePosition_col[0][0] = 'rePosition_magnets'
    rePosition_col = np.transpose(rePosition_col)
    print(rePosition_col)

    print(rePosition_col.shape)
    print(robotFilearray.shape)

    # print(np.append(robotFilearray, rePosition_col, axis=1))
    robotFilearray = np.hstack((robotFilearray[:, :8], rePosition_col, robotFilearray[:, 8:]))
    print(robotFilearray)

    # CSV file for the robot
    # with open('GAMA_'+batch+'/Output_for_Robot/Robot_GAMA_'+batch+'_tile_%03d.txt' % (tileNum),'w+') as robotFile:
    with open(robotFile, 'w+') as robotFile:
        writer = csv.writer(robotFile, delimiter=' ')
        writer.writerows(robotFilearray)

    return positioning_array,robotFilearray

def positioningArray_adjust_and_mergetoFile(positioning_array, plate_file, outputFile, newrow,newrow_circular):

    positioning_array_circular = np.vsplit(positioning_array, 2)[0]
    positioning_array_circular = np.insert(positioning_array_circular, 0, np.array(newrow_circular), 0)
    positioning_array_circular = np.delete(positioning_array_circular, [2, 3, 4, 5, 9, 10], 1)

    positioning_array = np.vsplit(positioning_array, 2)[1]
    positioning_array = np.insert(positioning_array, 0, np.array(newrow), 0)
    positioning_array = np.delete(positioning_array, [2, 3, 4, 5, 8], 1)

    index = 0
    # Open the input_file in read mode and output_file in write mode
    with open(plate_file, 'r') as read_obj, \
            open(outputFile, 'w+', newline='') as write_obj:
    # with open('GAMA_'+batch+'/Output/Hexa_and_Guides_GAMA_'+batch+'_tile_%03d.txt' % (tileNum), 'r') as read_obj, \
    #         open('GAMA_'+batch+'/Output_with_Positioning_array/Hexa_and_Guides_with_PositioningArray_GAMA_'+batch+'_tile_%03d.txt' % (tileNum), 'w+', newline='') as write_obj:
        # Create a csv.reader object from the input file object
        csv_reader = csv.reader(read_obj)
        # Create a csv.writer object from the output file object
        csv_writer = csv.writer(write_obj)
        # Read each row of the input csv file as list
        for row in csv_reader:
            roww_circular = np.array2string(positioning_array_circular[index], separator=' ',formatter={'str_kind': lambda x: x})
            roww = np.array2string(positioning_array[index], separator=' ', formatter={'str_kind': lambda x: x})
            # Append the default text in the row / list
            row.append(str(roww).replace('[', ' ').rstrip(']'))
            row.append(str(roww_circular).replace('[', ' ').rstrip(']'))
            # Add the updated row / list to the output file
            csv_writer.writerow(row)
            # index = index+32
            index += 1
            # open('galaxy_fields/output_1.txt', 'w').writelines([' '.join(row)])

    return positioning_array,positioning_array_circular

def finalFiles(outputFile, robotFile):

    df3 = pd.read_csv(outputFile, header=0)
    df3.to_csv(outputFile, index=False, sep=' ', quoting=csv.QUOTE_NONE, escapechar=' ')

    # df4 = pd.read_csv(robotFile, header=0, error_bad_lines=False)
    # df4.to_csv(robotFile, index=False, sep=' ', quoting=csv.QUOTE_NONE, escapechar=' ')

    # df3 = pd.read_csv('GAMA_'+batch+'/Output_with_Positioning_array/Hexa_and_Guides_with_PositioningArray_GAMA_'+batch+'_tile_%03d.txt' % (tileNum), header=0)
    # df3.to_csv('GAMA_'+batch+'/Output_with_Positioning_array/Hexa_and_Guides_with_PositioningArray_GAMA_'+batch+'_tile_%03d.txt' % (tileNum), \
    #             index=False, sep=' ', quoting=csv.QUOTE_NONE, escapechar=' ')
    # df4 = pd.read_csv('GAMA_' + batch + '/Output_for_Robot/Robot_GAMA_' + batch + '_tile_%03d.txt' % (tileNum),header=0)
    # df4.to_csv('GAMA_' + batch + '/Output_for_Robot/Robot_GAMA_' + batch + '_tile_%03d.txt' % (tileNum),\
    #            index=False, sep=' ', quoting=csv.QUOTE_NONE, escapechar=' ')




def intersection_slope_segment_and_vertical_segment(vertical_segment, slope_segment):

    line = slope_segment.create_line_equation()
    intersection_point_y = line.calculate_y_from_x(vertical_segment.start_point[0])
    intersection_point = [vertical_segment.start_point[0],intersection_point_y]

    if vertical_segment.point_is_on_segment(intersection_point):

        if slope_segment.point_is_on_segment(intersection_point):

            return intersection_point

        else:

            return None
    else:

        return None

def segments_intersection(segment1,segment2):

    intersection_point =[]

    if segment1.vertical():

        if segment2.vertical():

            if segment1.start_point[0] == segment2.start_point[0]:
                ### add check if point same segments

                intersection_point = [segment2.start_point[0], None]
        else:

            intersection_point = intersection_slope_segment_and_vertical_segment(segment1,segment2)

    elif segment2.vertical():

        intersection_point = intersection_slope_segment_and_vertical_segment(segment2,segment1)

    else:
        line1 = segment1.create_line_equation()
        line2 = segment2.create_line_equation()

        if line1.slope != line2.slope:

            intersection_point_x = (line2.origin - line1.origin) / (line1.slope - line2.slope)
            intersection_point_y = line1.slope * intersection_point_x + line1.origin

            if segment1.point_is_on_segment([intersection_point_x,intersection_point_y]):
                if segment2.point_is_on_segment([intersection_point_x,intersection_point_y]):

                    intersection_point = [intersection_point_x,intersection_point_y]

    return intersection_point



import matplotlib.pyplot as plt
from ..general_operations.geometry_shapes.circle import circle
from ..hector.constants import HECTOR_plate_center_coordinate,HECTOR_plate_radius,robot_center_x,robot_center_y

class HECTOR_plate(circle):

    def __init__(self):
        self.center = HECTOR_plate_center_coordinate
        self.radius = HECTOR_plate_radius
        self.orientation = 0

def plot_plate(HECTOR_plate):

    # # setting aspect ratio
    # plt.figure()
    # plt.plot(np.sin(np.linspace(0, 20, 100)))
    # plt.figure().canvas.layout.width = '800px'
    # plt.figure().canvas.draw()

    plt.plot(HECTOR_plate[0], HECTOR_plate[1], 'r')
    plt.plot(HECTOR_plate[0], HECTOR_plate[2], 'r')
    #plt.plot(0, 0, 'x')

    plt.axis('scaled')
    # plt.set_axis(xlim=(-350, 350), ylim=(-350, 350))
    # plt.gca().set_xlim([-400, 400])
    # plt.gca().set_ylim([-400, 400])


    plt.xlabel('x (mm)')
    plt.ylabel('y (mm)')
    #plt.show()

###############################################################
    #2nd figure
    plt.figure()

    plt.plot(HECTOR_plate[0], HECTOR_plate[1], 'r')
    plt.plot(HECTOR_plate[0], HECTOR_plate[2], 'r')
    # plt.plot(0, 0, 'x')

    plt.axis('scaled')
    plt.axis([robot_center_x - 260, robot_center_x + 260, robot_center_y + 260, robot_center_y - 260])
    plt.xlabel('x (mm)')
    plt.ylabel('y (mm)')
    #plt.show()
###############################################################



from setuptools import setup, find_packages, Command
from setuptools.command.install import install
import distutils.command.install as orig
import subprocess
import inspect
import pathlib

class CustomInstall(install):

    user_options = install.user_options

    def initialize_options(self):
        install.initialize_options(self)

    def finalize_options(self):
        install.finalize_options(self)

    def run(self):

        # Copied from the original setup.command.install functions
        if not install._called_from_setup(inspect.currentframe()):
            # Run in backward-compatibility mode to support bdist_* commands.
            orig.install.run(self)
        else:
            install.do_egg_install()

        # and now adding in my own bit
        print("...building the Hector Translation code")
        subprocess.call(
            ['make', 
            '-C', 
            'hop/distortion_correction/HectorTranslationSoftware/Code']
        )
        print("...Done!")




# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()


setup(name='Hector-Observations-Pipeline',
      version='0.2.3',
      description='Hector Galaxy Survey Observations pipeline',
      long_description=README,
      long_description_content_type="text/markdown",
      author='Sam Vaughan',
      author_email='sam.vaughan@sydney.edu.au',
      url='https://github.com/samvaughan/Hector-Observations-Pipeline',
      license="GNU",
      classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8"],
      packages=find_packages(),
      include_package_data = True,
      cmdclass={'install': CustomInstall}
     )




import numpy as np
import matplotlib.pyplot as plt
from ...hector.constants import robot_center_x,robot_center_y

class circle:

    def __init__(self, center,radius,orientation):

        self.center = center
        self.radius = radius
        self.orientation = orientation

        # fixing plotting view for Robot file coordinates
        self.view_y = self.center[0]
        self.view_x = - self.center[1]
        # self.view_orientation = self.orientation# - 90

    def calculate_circle_all_coordinates(self):

        self.x_coordinates = np.arange(self.center[0] - self.radius, self.center[0] + self.radius, self.radius / 100000)
        self.y_coordinates_negative = -np.sqrt(np.abs(self.radius ** 2 - (self.x_coordinates - self.center[0]) ** 2)) + self.center[1]
        self.y_coordinates_positive = np.sqrt(np.abs(self.radius ** 2 - (self.x_coordinates - self.center[0]) ** 2)) + self.center[1]


    def draw_circle(self,colour):

        circle.calculate_circle_all_coordinates(self)

        plt.figure(1)
        plt.plot(self.x_coordinates, self.y_coordinates_negative, colour, alpha = 0.5)
        plt.plot(self.x_coordinates, self.y_coordinates_positive, colour, alpha = 0.5)
        plt.axis('scaled')

################################################################
        plt.figure(2)
        plt.plot(self.x_coordinates + robot_center_x, (self.y_coordinates_negative + robot_center_y), colour)
        plt.plot(self.x_coordinates + robot_center_x, (self.y_coordinates_positive + robot_center_y), colour)
        plt.axis('scaled')
################################################################



from .conflicts.functions import create_list_of_blocking_magnets,\
                                all_blocking_magnets_are_fully_blocked
from ..general_operations.trigonometry import convert_radians_to_degrees
from .hexabundle import create_Magnet_Dictionary, get_probe_count, star_hexabundle_allocation, \
                                cores91_hexabundles_sortedByMstar, resolve_hexabundle_allocation_repeats, \
                                resolve_unallocated_hexabundle, check_for_hexabundle_allocation_repeats, \
                                check_for_unallocated_hexabundle, update_galaxyIDrecord_withHexabundle, \
                                hexabundle_allocation_fromMagnetDict, overall_hexabundle_size_allocation_operation_version1, \
                                overall_hexabundle_size_allocation_operation_version2_median, \
                                overall_hexabundle_size_allocation_operation_version3_largerBundlePriority
from ..problem_operations.robot_parameters import assign_magnet_labels, add_pickupDirection, \
                                                assign_preferable_pickupDirection, calculate_pickup_putdown_angles, \
                                                check_for_negative_twopi_angle, check_for_unresolvable_conflict
from ..hector.constants import robot_center_x,robot_center_y
import numpy as np
import string
import sys
import re

def calculate_placement_ordering_of_blocked_magnet(blocked_magnet,list_of_conflicts,list_of_fully_blocked_magnets):

    blocking_magnets = create_list_of_blocking_magnets(list_of_conflicts, blocked_magnet)

    for i in range(50):

        if not all_blocking_magnets_are_fully_blocked(blocking_magnets,list_of_fully_blocked_magnets):

            blocked_magnet.placement_index += 1

            break

        else:

            blocking_magnets = []

            for magnet in blocking_magnets:
                blocking_magnets.append(create_list_of_blocking_magnets(list_of_conflicts, magnet))


    if i == 49:
        blocked_magnet.placement_index = None
        print('Error ! ',blocked_magnet.__class__.__name__,int(blocked_magnet.index),' cannot be placed')



def calculate_placement_ordering_of_all_blocked_magnets(list_of_fully_blocked_magnets,list_of_conflicts):

    for blocked_magnet in list_of_fully_blocked_magnets:
        calculate_placement_ordering_of_blocked_magnet(blocked_magnet,list_of_conflicts,list_of_fully_blocked_magnets)


def create_position_ordering_array(all_magnets, fully_blocked_magnets, conflicted_magnets, galaxyIDrecord, \
                                   clusterNum, tileNum, conflictFile, flagsFile):

    calculate_placement_ordering_of_all_blocked_magnets(fully_blocked_magnets, conflicted_magnets)

    max_order = max(magnet.placement_index for magnet in all_magnets if magnet.placement_index is not None)

    position_ordering_array = []

    # magnet label and guide hexabundle index
    index1 = 1
    guideIndex = 1

    # carrying out the whole hexabundle allocation algorithm from hexabundles.py script
    galaxyIDrecord, MagnetDict = overall_hexabundle_size_allocation_operation_version3_largerBundlePriority(all_magnets, \
                                                                        galaxyIDrecord, clusterNum, tileNum, flagsFile)

    for magnet in all_magnets:

        # identifying possible pickup directions available for magnet
        available_pickup = [area.code for area in magnet.pickup_areas]

        # checking for magnets with unresolvable conflicts, which are flagged and assigned order of '0'
        order = check_for_unresolvable_conflict(magnet,conflictFile,max_order,clusterNum,tileNum)

        # adding magnet labels of rectangular: R01.. ,and circular: Blu,Mag,Gre,Yel
        magnet,index1 = assign_magnet_labels(magnet,index1)

        # adding pickup direction to placement orders of 1 in the case of more than 1 placement orders
        magnet, available_pickup = add_pickupDirection(magnet, available_pickup)

        # selecting only 1 pickup direction based on preference for circular: TR,TL,RO,RI and rectangular: I,O
        available_pickup = assign_preferable_pickupDirection(available_pickup)

        # assigning allocated hexabundles to respective magnet as stored in Magnet Dictionary
        magnet, guideIndex = hexabundle_allocation_fromMagnetDict (MagnetDict,magnet,guideIndex)

        # calculating rotation angles for pickup and putdown based on available pickup direction
        magnet = calculate_pickup_putdown_angles(magnet,available_pickup)

        # checking rotation angle to ensure it is within range of 0 to 360 degrees
        magnet.rotation_putdown = check_for_negative_twopi_angle(magnet.rotation_putdown)

        ## ** Might need to change center coordinates of magnets to different scale for considering optical and other type of distortion **

        # storing all the parameters in positioning array
        f = np.append([magnet.__class__.__name__, str(magnet.magnet_label), str(robot_center_x+magnet.view_x), \
                       str(robot_center_y+magnet.view_y), str(magnet.rotation_pickup), str(round(magnet.rotation_putdown,2)), \
                       str(order), available_pickup, str(float(magnet.IDs)), str(int(magnet.index))], str(magnet.hexabundle))
        position_ordering_array.append(np.array(f))


    return np.array(position_ordering_array),galaxyIDrecord


